@@ -12,32 +12,37 @@
 /* .in -4
 /*	} BOUNCE_INFO;
 /*
-/*	BOUNCE_INFO *bounce_mail_init(service, queue_name, queue_id,
-/*					encoding, dsn_envid, template)
+/*	BOUNCE_INFO *bounce_mail_init(service, queue_name, queue_id, encoding,
+/*					sendopts, dsn_envid, template)
 /*	const char *service;
 /*	const char *queue_name;
 /*	const char *queue_id;
 /*	const char *encoding;
+/*	int	sendopts;
 /*	const char *dsn_envid;
 /*	const BOUNCE_TEMPLATE *template;
 /*
 /*	BOUNCE_INFO *bounce_mail_one_init(queue_name, queue_id, encoding,
-/*					dsn_envid, dsn_notify, rcpt, dsn)
+/*					sendopts, dsn_envid, dsn_notify,
+/*					rcpt_buf, dsn_buf, template)
 /*	const char *queue_name;
 /*	const char *queue_id;
 /*	const char *encoding;
+/*	int	sendopts;
 /*	int	dsn_notify;
 /*	const char *dsn_envid;
-/*	RECIPIENT *rcpt;
-/*	DSN	*dsn;
+/*	RCPT_BUF *rcpt_buf;
+/*	DSN_BUF	*dsn_buf;
+/*	const BOUNCE_TEMPLATE *template;
 /*
 /*	void	bounce_mail_free(bounce_info)
 /*	BOUNCE_INFO *bounce_info;
 /*
-/*	int	bounce_header(fp, bounce_info, recipient)
+/*	int	bounce_header(fp, bounce_info, recipient, postmaster_copy)
 /*	VSTREAM *fp;
 /*	BOUNCE_INFO *bounce_info;
 /*	const char *recipient;
+/*	int	postmaster_copy;
 /*
 /*	int	bounce_boilerplate(fp, bounce_info)
 /*	VSTREAM *fp;
@@ -86,15 +91,15 @@
 /*	undeliverable message.
 /*
 /*	bounce_mail_one_init() provides the same function for only
-/*	one recipient that is not read from bounce logfile. It
-/*	assumes a template type of FAIL_TEMPLATE().
+/*	one recipient that is not read from bounce logfile.
 /*
 /*	bounce_mail_free() releases memory allocated by bounce_mail_init()
 /*	and closes any files opened by bounce_mail_init().
 /*
 /*	bounce_header() produces a standard mail header with the specified
 /*	recipient and starts a text/plain message segment for the
-/*	human-readable problem description.
+/*	human-readable problem description. postmaster_copy is either
+/*	POSTMASTER_COPY or NO_POSTMASTER_COPY.
 /*
 /*	bounce_boilerplate() produces the standard "sorry" text that
 /*	creates the illusion that mail systems are civilized.
@@ -150,6 +155,14 @@
 /*	IBM T.J. Watson Research
 /*	P.O. Box 704
 /*	Yorktown Heights, NY 10598, USA
+/*
+/*	Wietse Venema
+/*	Google, Inc.
+/*	111 8th Avenue
+/*	New York, NY 10011, USA
+/*
+/*	Wietse Venema
+/*	porcupine.org
 /*--*/
 
 /* System library. */
@@ -157,6 +170,7 @@
 #include <sys_defs.h>
 #include <sys/stat.h>
 #include <stdlib.h>
+#include <stdio.h>			/* sscanf() */
 #include <unistd.h>
 #include <errno.h>
 #include <string.h>
@@ -194,12 +208,15 @@
 #include <lex_822.h>
 #include <deliver_completed.h>
 #include <dsn_mask.h>
+#include <smtputf8.h>
+#include <header_opts.h>
 
 /* Application-specific. */
 
 #include "bounce_service.h"
 
 #define STR vstring_str
+#define LEN VSTRING_LEN
 
 /* bounce_mail_alloc - initialize */
 
@@ -207,22 +224,35 @@ static BOUNCE_INFO *bounce_mail_alloc(const char *service,
 				              const char *queue_name,
 				              const char *queue_id,
 				              const char *encoding,
+				              int sendopts,
 				              const char *dsn_envid,
-				            const BOUNCE_TEMPLATE *template,
+				              RCPT_BUF *rcpt_buf,
+				              DSN_BUF *dsn_buf,
+				              BOUNCE_TEMPLATE *template,
 				              BOUNCE_LOG *log_handle)
 {
     BOUNCE_INFO *bounce_info;
     int     rec_type;
+    int     skip_message_segment = 0;
 
     /*
-     * Bundle up a bunch of parameters and initialize information. that will
+     * Bundle up a bunch of parameters and initialize information that will
      * be discovered on the fly.
+     * 
+     * XXX Instead of overriding the returned-message MIME encoding, separate
+     * the returned-message MIME encoding from the (boiler plate, delivery
+     * status) MIME encoding.
      */
     bounce_info = (BOUNCE_INFO *) mymalloc(sizeof(*bounce_info));
     bounce_info->service = service;
     bounce_info->queue_name = queue_name;
     bounce_info->queue_id = queue_id;
-    if (strcmp(encoding, MAIL_ATTR_ENC_8BIT) == 0) {
+    bounce_info->sendopts = sendopts;
+    /* Fix 20140708: override MIME encoding: addresses may be 8bit. */
+    /* Fix 20140718: override MIME encoding: 8bit $myhostname expansion. */
+    if (var_smtputf8_enable /* was: bounce_info->smtputf8 */ ) {
+	bounce_info->mime_encoding = "8bit";
+    } else if (strcmp(encoding, MAIL_ATTR_ENC_8BIT) == 0) {
 	bounce_info->mime_encoding = "8bit";
     } else if (strcmp(encoding, MAIL_ATTR_ENC_7BIT) == 0) {
 	bounce_info->mime_encoding = "7bit";
@@ -242,6 +272,9 @@ static BOUNCE_INFO *bounce_mail_alloc(const char *service,
     bounce_info->arrival_time = 0;
     bounce_info->orig_offs = 0;
     bounce_info->message_size = 0;
+    bounce_info->orig_msgid = vstring_alloc(100);
+    bounce_info->rcpt_buf = rcpt_buf;
+    bounce_info->dsn_buf = dsn_buf;
     bounce_info->log_handle = log_handle;
 
     /*
@@ -275,9 +308,9 @@ static BOUNCE_INFO *bounce_mail_alloc(const char *service,
 	msg_fatal("open %s %s: %m", service, queue_id);
 
     /*
-     * Skip over the original message envelope records. If the envelope is
-     * corrupted just send whatever we can (remember this is a best effort,
-     * it does not have to be perfect).
+     * Get time/size/sender information from the original message envelope
+     * records. If the envelope is corrupted just send whatever we can
+     * (remember this is a best effort, it does not have to be perfect).
      * 
      * Lock the file for shared use, so that queue manager leaves it alone after
      * restarting.
@@ -289,29 +322,98 @@ static BOUNCE_INFO *bounce_mail_alloc(const char *service,
 		    DELIVER_LOCK_MODE) < 0)
 	    msg_fatal("cannot get shared lock on %s: %m",
 		      VSTREAM_PATH(bounce_info->orig_fp));
-	while ((rec_type = rec_get(bounce_info->orig_fp,
-				   bounce_info->buf, 0)) > 0) {
+	while ((rec_type =
+		rec_get(bounce_info->orig_fp, bounce_info->buf, 0)) > 0) {
+
+	    /*
+	     * Postfix version dependent: data offset in SIZE record.
+	     */
 	    if (rec_type == REC_TYPE_SIZE) {
-		if (bounce_info->message_size == 0
-		    && (bounce_info->message_size = atol(STR(bounce_info->buf))) < 0)
+		if (bounce_info->message_size == 0)
+		    sscanf(STR(bounce_info->buf), "%ld %ld",
+			   &bounce_info->message_size,
+			   &bounce_info->orig_offs);
+		if (bounce_info->message_size < 0)
 		    bounce_info->message_size = 0;
-	    } else if (rec_type == REC_TYPE_TIME) {
+		if (bounce_info->orig_offs < 0)
+		    bounce_info->orig_offs = 0;
+	    }
+
+	    /*
+	     * Information for the Arrival-Date: attribute.
+	     */
+	    else if (rec_type == REC_TYPE_TIME) {
 		if (bounce_info->arrival_time == 0
 		    && (bounce_info->arrival_time = atol(STR(bounce_info->buf))) < 0)
 		    bounce_info->arrival_time = 0;
-	    } else if (rec_type == REC_TYPE_FROM) {
+	    }
+
+	    /*
+	     * Information for the X-Postfix-Sender: attribute.
+	     */
+	    else if (rec_type == REC_TYPE_FROM) {
 		quote_822_local_flags(bounce_info->sender,
 				      VSTRING_LEN(bounce_info->buf) ?
 				      STR(bounce_info->buf) :
 				      mail_addr_mail_daemon(), 0);
-	    } else if (rec_type == REC_TYPE_MESG) {
+	    }
+
+	    /*
+	     * Backwards compatibility: no data offset in SIZE record.
+	     */
+	    else if (rec_type == REC_TYPE_MESG) {
 		/* XXX Future: sender+recipient after message content. */
 		if (VSTRING_LEN(bounce_info->sender) == 0)
 		    msg_warn("%s: no sender before message content record",
 			     bounce_info->queue_id);
 		bounce_info->orig_offs = vstream_ftell(bounce_info->orig_fp);
+		skip_message_segment = 1;
+	    }
+
+	    /*
+	     * Extract Message-ID from extracted segment, for use in threaded
+	     * bounces.
+	     */
+	    else if (rec_type == REC_TYPE_ATTR && var_threaded_bounce) {
+		char   *cp = STR(bounce_info->buf);
+		ssize_t len = sizeof(MAIL_ATTR_MESSAGE_ID);
+		char   *err;
+
+		if (strncmp(cp, MAIL_ATTR_MESSAGE_ID "=", len) == 0) {
+		    cp += len;
+		    if ((err = extpar(&cp, "<>", EXTPAR_FLAG_NONE)) != 0) {
+			msg_warn("%s: malformed Message-ID attribute: %s",
+				 bounce_info->queue_id, err);
+			myfree(err);
+		    } else {
+			vstring_sprintf(bounce_info->orig_msgid, "<%s>", cp);
+		    }
+		}
+	    }
+
+	    /*
+	     * Are we done yet?
+	     */
+	    if (bounce_info->orig_offs > 0
+		&& bounce_info->arrival_time > 0
+		&& VSTRING_LEN(bounce_info->sender) > 0
+		&& (var_threaded_bounce == 0
+		    || VSTRING_LEN(bounce_info->orig_msgid) > 0)) {
 		break;
 	    }
+
+	    /*
+	     * Skip over (the remainder of) the message segment. If that
+	     * fails, degrade.
+	     */
+	    if (skip_message_segment) {
+		if (vstream_fseek(bounce_info->orig_fp,
+				  bounce_info->orig_offs +
+				  bounce_info->message_size,
+				  SEEK_SET) < 0)
+		     /* void */ ;
+		skip_message_segment = 0;
+	    }
 	}
     }
     return (bounce_info);
@@ -323,11 +425,14 @@ BOUNCE_INFO *bounce_mail_init(const char *service,
 			              const char *queue_name,
 			              const char *queue_id,
 			              const char *encoding,
+			              int sendopts,
 			              const char *dsn_envid,
-			              const BOUNCE_TEMPLATE *template)
+			              BOUNCE_TEMPLATE *template)
 {
     BOUNCE_INFO *bounce_info;
     BOUNCE_LOG *log_handle;
+    RCPT_BUF *rcpt_buf;
+    DSN_BUF *dsn_buf;
 
     /*
      * Initialize the bounce_info structure. If the bounce log cannot be
@@ -337,11 +442,18 @@ BOUNCE_INFO *bounce_mail_init(const char *service,
      * job. But if the system IS running out of resources, raise a fatal
      * run-time error and force a backoff.
      */
-    if ((log_handle = bounce_log_open(service, queue_id, O_RDONLY, 0)) == 0
-	&& errno != ENOENT)
-	msg_fatal("open %s %s: %m", service, queue_id);
+    if ((log_handle = bounce_log_open(service, queue_id, O_RDONLY, 0)) == 0) {
+	if (errno != ENOENT)
+	    msg_fatal("open %s %s: %m", service, queue_id);
+	rcpt_buf = 0;
+	dsn_buf = 0;
+    } else {
+	rcpt_buf = rcpb_create();
+	dsn_buf = dsb_create();
+    }
     bounce_info = bounce_mail_alloc(service, queue_name, queue_id, encoding,
-				    dsn_envid, template, log_handle);
+				    sendopts, dsn_envid, rcpt_buf, dsn_buf,
+				    template, log_handle);
     return (bounce_info);
 }
 
@@ -350,20 +462,20 @@ BOUNCE_INFO *bounce_mail_init(const char *service,
 BOUNCE_INFO *bounce_mail_one_init(const char *queue_name,
 				          const char *queue_id,
 				          const char *encoding,
+				          int sendopts,
 				          const char *dsn_envid,
-				          RECIPIENT *rcpt,
-				          DSN *dsn)
+				          RCPT_BUF *rcpt_buf,
+				          DSN_BUF *dsn_buf,
+				          BOUNCE_TEMPLATE *template)
 {
     BOUNCE_INFO *bounce_info;
-    BOUNCE_LOG *log_handle;
 
     /*
-     * Initialize the bounce_info structure. Forge a logfile record for just
-     * one recipient.
+     * Initialize the bounce_info structure for just one recipient.
      */
-    log_handle = bounce_log_forge(rcpt, dsn);
     bounce_info = bounce_mail_alloc("none", queue_name, queue_id, encoding,
-				    dsn_envid, FAIL_TEMPLATE(), log_handle);
+				    sendopts, dsn_envid, rcpt_buf, dsn_buf,
+				    template, (BOUNCE_LOG *) 0);
     return (bounce_info);
 }
 
@@ -371,9 +483,14 @@ BOUNCE_INFO *bounce_mail_one_init(const char *queue_name,
 
 void    bounce_mail_free(BOUNCE_INFO *bounce_info)
 {
-    if (bounce_info->log_handle && bounce_log_close(bounce_info->log_handle))
-	msg_warn("%s: read bounce log %s: %m",
-		 bounce_info->queue_id, bounce_info->queue_id);
+    if (bounce_info->log_handle) {
+	if (bounce_log_close(bounce_info->log_handle))
+	    msg_warn("%s: read bounce log %s: %m",
+		     bounce_info->queue_id, bounce_info->queue_id);
+	vstring_free(bounce_info->orig_msgid);
+	rcpb_free(bounce_info->rcpt_buf);
+	dsb_free(bounce_info->dsn_buf);
+    }
     if (bounce_info->orig_fp && vstream_fclose(bounce_info->orig_fp))
 	msg_warn("%s: read message file %s %s: %m",
 		 bounce_info->queue_id, bounce_info->queue_name,
@@ -381,40 +498,45 @@ void    bounce_mail_free(BOUNCE_INFO *bounce_info)
     vstring_free(bounce_info->buf);
     vstring_free(bounce_info->sender);
     myfree(bounce_info->mail_name);
-    myfree((char *) bounce_info->mime_boundary);
-    myfree((char *) bounce_info);
+    myfree((void *) bounce_info->mime_boundary);
+    myfree((void *) bounce_info);
 }
 
 /* bounce_header - generate bounce message header */
 
 int     bounce_header(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
-		              const char *dest)
+		              const char *dest, int postmaster_copy)
 {
-    const BOUNCE_TEMPLATE *template = bounce_info->template;
-    int     postmaster_copy;
+    BOUNCE_TEMPLATE *template = bounce_info->template;
 
     /*
      * Print a minimal bounce header. The cleanup service will add other
      * headers and will make all addresses fully qualified.
      */
 #define STREQ(a, b) (strcasecmp((a), (b)) == 0)
-
-    /*
-     * XXX This should be caller specified.
-     */
-    postmaster_copy =
-	(template->postmaster_subject != 0
-	 && (dest == var_bounce_rcpt || dest == var_2bounce_rcpt
-	     || dest == var_delay_rcpt));
+#define STRNE(a, b) (strcasecmp((a), (b)) != 0)
 
     /*
      * Generic headers.
      */
-    post_mail_fprintf(bounce, "From: %s", template->from);
-    post_mail_fprintf(bounce, "Subject: %s", postmaster_copy ?
-		      template->postmaster_subject : template->subject);
-    post_mail_fprintf(bounce, "To: %s",
-		      STR(quote_822_local(bounce_info->buf, dest)));
+    bounce_template_headers(post_mail_fprintf, bounce, template,
+			    STR(quote_822_local(bounce_info->buf, dest)),
+			    postmaster_copy);
+
+    /*
+     * References and Reply-To header that references the original message-id
+     * for better threading in MUAs.
+     */
+    if (VSTRING_LEN(bounce_info->orig_msgid) > 0) {
+	post_mail_fprintf(bounce, "References: %s", STR(bounce_info->orig_msgid));
+	post_mail_fprintf(bounce, "In-Reply-To: %s", STR(bounce_info->orig_msgid));
+    }
+
+    /*
+     * Auto-Submitted header, as per RFC 3834.
+     */
+    post_mail_fprintf(bounce, "Auto-Submitted: %s", postmaster_copy ?
+		      "auto-generated" : "auto-replied");
 
     /*
      * MIME header. Use 8bit encoding when either the bounced message or the
@@ -427,18 +549,33 @@ int     bounce_header(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
     if (bounce_info->mime_encoding)
 	post_mail_fprintf(bounce, "Content-Transfer-Encoding: %s",
 		     STREQ(bounce_info->mime_encoding, MAIL_ATTR_ENC_7BIT) ?
-		      template->mime_encoding : bounce_info->mime_encoding);
+			  bounce_template_encoding(template) :
+			  bounce_info->mime_encoding);
     post_mail_fputs(bounce, "");
     post_mail_fputs(bounce, "This is a MIME-encapsulated message.");
 
     /*
      * MIME header.
      */
+#define NOT_US_ASCII(tp) \
+	STRNE(bounce_template_charset(template), "us-ascii")
+
+#define NOT_7BIT_MIME(bp) \
+	(bp->mime_encoding && STRNE(bp->mime_encoding, MAIL_ATTR_ENC_7BIT))
+
     post_mail_fputs(bounce, "");
     post_mail_fprintf(bounce, "--%s", bounce_info->mime_boundary);
     post_mail_fprintf(bounce, "Content-Description: %s", "Notification");
+    /* Fix 20140718: UTF-8 address or $myhostname expansion. */
     post_mail_fprintf(bounce, "Content-Type: %s; charset=%s",
-		      "text/plain", template->charset);
+		      "text/plain", NOT_US_ASCII(template) ?
+		      bounce_template_charset(template) :
+		      NOT_7BIT_MIME(bounce_info) ?
+		      "utf-8" : "us-ascii");
+    /* Fix 20140709: addresses may be 8bit. */
+    if (NOT_7BIT_MIME(bounce_info))
+	post_mail_fprintf(bounce, "Content-Transfer-Encoding: %s",
+			  bounce_info->mime_encoding);
     post_mail_fputs(bounce, "");
 
     return (vstream_ferror(bounce));
@@ -458,7 +595,7 @@ int     bounce_boilerplate(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
 
 /* bounce_print - line_wrap callback */
 
-static void bounce_print(const char *str, int len, int indent, char *context)
+static void bounce_print(const char *str, int len, int indent, void *context)
 {
     VSTREAM *bounce = (VSTREAM *) context;
 
@@ -479,13 +616,15 @@ static void bounce_print_wrap(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
     vstring_vsprintf(bounce_info->buf, format, ap);
     va_end(ap);
     line_wrap(STR(bounce_info->buf), LENGTH, INDENT,
-	      bounce_print, (char *) bounce);
+	      bounce_print, (void *) bounce);
 }
 
 /* bounce_recipient_log - send one bounce log report entry */
 
 int     bounce_recipient_log(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
 {
+    RECIPIENT *rcpt = &bounce_info->rcpt_buf->rcpt;
+    DSN    *dsn = &bounce_info->dsn_buf->dsn;
 
     /*
      * Mask control and non-ASCII characters (done in bounce_log_read()),
@@ -493,16 +632,15 @@ int     bounce_recipient_log(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
      * piped into other programs. Sort of like TCP Wrapper's safe_finger
      * program.
      */
+#define NON_NULL_EMPTY(s) ((s) && *(s))
+
     post_mail_fputs(bounce, "");
-    if (bounce_info->log_handle->rcpt.orig_addr) {
+    if (NON_NULL_EMPTY(rcpt->orig_addr)) {
 	bounce_print_wrap(bounce, bounce_info, "<%s> (expanded from <%s>): %s",
-			  bounce_info->log_handle->rcpt.address,
-			  bounce_info->log_handle->rcpt.orig_addr,
-			  bounce_info->log_handle->dsn.reason);
+			  rcpt->address, rcpt->orig_addr, dsn->reason);
     } else {
 	bounce_print_wrap(bounce, bounce_info, "<%s>: %s",
-			  bounce_info->log_handle->rcpt.address,
-			  bounce_info->log_handle->dsn.reason);
+			  rcpt->address, dsn->reason);
     }
     return (vstream_ferror(bounce));
 }
@@ -512,6 +650,7 @@ int     bounce_recipient_log(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
 int     bounce_diagnostic_log(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
 			              int notify_filter)
 {
+    RECIPIENT *rcpt = &bounce_info->rcpt_buf->rcpt;
     int     count = 0;
 
     /*
@@ -525,14 +664,16 @@ int     bounce_diagnostic_log(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
      */
     if (bounce_info->log_handle == 0
 	|| bounce_log_rewind(bounce_info->log_handle)) {
-	if (IS_FAIL_TEMPLATE(bounce_info->template)) {
+	if (IS_FAILURE_TEMPLATE(bounce_info->template)) {
+	    post_mail_fputs(bounce, "");
 	    post_mail_fputs(bounce, "\t--- Delivery report unavailable ---");
 	    count = 1;				/* XXX don't abort */
 	}
     } else {
-	while (bounce_log_read(bounce_info->log_handle) != 0) {
-	    if (bounce_info->log_handle->rcpt.dsn_notify == 0	/* compat */
-	    || (bounce_info->log_handle->rcpt.dsn_notify & notify_filter)) {
+	while (bounce_log_read(bounce_info->log_handle, bounce_info->rcpt_buf,
+			       bounce_info->dsn_buf) != 0) {
+	    if (rcpt->dsn_notify == 0		/* compat */
+		|| (rcpt->dsn_notify & notify_filter)) {
 		count++;
 		if (bounce_recipient_log(bounce, bounce_info) != 0)
 		    break;
@@ -554,7 +695,16 @@ int     bounce_header_dsn(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
     post_mail_fprintf(bounce, "--%s", bounce_info->mime_boundary);
     post_mail_fprintf(bounce, "Content-Description: %s",
 		      "Delivery report");
-    post_mail_fprintf(bounce, "Content-Type: %s", "message/delivery-status");
+    /* Generate *global* only if the original requested SMTPUTF8 support. */
+    post_mail_fprintf(bounce, "Content-Type: message/%sdelivery-status",
+		      (bounce_info->sendopts & SMTPUTF8_FLAG_REQUESTED) ?
+		      "global-" : "");
+    /* Fix 20140709: addresses may be 8bit. */
+    if (NOT_7BIT_MIME(bounce_info)
+    /* BC Fix 20170610: prevent MIME downgrade of message/delivery-status. */
+	&& (bounce_info->sendopts & SMTPUTF8_FLAG_REQUESTED))
+	post_mail_fprintf(bounce, "Content-Transfer-Encoding: %s",
+			  bounce_info->mime_encoding);
 
     /*
      * According to RFC 1894: The body of a message/delivery-status consists
@@ -567,15 +717,23 @@ int     bounce_header_dsn(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
 #if 0
     post_mail_fprintf(bounce, "Received-From-MTA: dns; %s", "whatever");
 #endif
-    if (bounce_info->dsn_envid) {
+    if (NON_NULL_EMPTY(bounce_info->dsn_envid)) {
 	post_mail_fprintf(bounce, "Original-Envelope-Id: %s",
 			  bounce_info->dsn_envid);
     }
     post_mail_fprintf(bounce, "X-%s-Queue-ID: %s",
 		      bounce_info->mail_name, bounce_info->queue_id);
+
+#define IS_UTF8_ADDRESS(str) \
+	((str)[0] != 0 && !allascii(str) && valid_utf8_stringz(str))
+
+    /* Fix 20140708: use "utf-8" or "rfc822" as appropriate. */
     if (VSTRING_LEN(bounce_info->sender) > 0)
-	post_mail_fprintf(bounce, "X-%s-Sender: rfc822; %s",
-			  bounce_info->mail_name, STR(bounce_info->sender));
+	post_mail_fprintf(bounce, "X-%s-Sender: %s; %s",
+			  bounce_info->mail_name,
+			  (bounce_info->sendopts & SMTPUTF8_FLAG_ALL)
+			  && IS_UTF8_ADDRESS(STR(bounce_info->sender)) ?
+			  "utf-8" : "rfc822", STR(bounce_info->sender));
     if (bounce_info->arrival_time > 0)
 	post_mail_fprintf(bounce, "Arrival-Date: %s",
 			  mail_date(bounce_info->arrival_time));
@@ -586,9 +744,15 @@ int     bounce_header_dsn(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
 
 int     bounce_recipient_dsn(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
 {
+    RECIPIENT *rcpt = &bounce_info->rcpt_buf->rcpt;
+    DSN    *dsn = &bounce_info->dsn_buf->dsn;
+
     post_mail_fputs(bounce, "");
-    post_mail_fprintf(bounce, "Final-Recipient: rfc822; %s",
-		      bounce_info->log_handle->rcpt.address);
+    /* Fix 20140708: Don't send "utf-8" type with non-UTF8 address. */
+    post_mail_fprintf(bounce, "Final-Recipient: %s; %s",
+		      (bounce_info->sendopts & SMTPUTF8_FLAG_ALL)
+		      && IS_UTF8_ADDRESS(rcpt->address) ?
+		      "utf-8" : "rfc822", rcpt->address);
 
     /*
      * XXX DSN
@@ -602,42 +766,39 @@ int     bounce_recipient_dsn(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
      * relayed.". Postfix adds an ORCPT parameter under these conditions.
      * 
      * Therefore, all down-stream MTAs will send DSNs with Original-Recipient
-     * field ontaining this same ORCPT value. When a down-stream MTA can use
+     * field containing this same ORCPT value. When a down-stream MTA can use
      * that information in their DSNs, it makes no sense that an up-stream
      * MTA can't use that same information in its own DSNs.
      * 
      * Postfix always reports an Original-Recipient field, because it is more
      * more useful and more consistent.
      */
-    if (bounce_info->log_handle->rcpt.dsn_orcpt) {
-	post_mail_fprintf(bounce, "Original-Recipient: %s",
-			  bounce_info->log_handle->rcpt.dsn_orcpt);
-    } else if (bounce_info->log_handle->rcpt.orig_addr) {
-	post_mail_fprintf(bounce, "Original-Recipient: rfc822; %s",
-			  bounce_info->log_handle->rcpt.orig_addr);
+    if (NON_NULL_EMPTY(rcpt->dsn_orcpt)) {
+	post_mail_fprintf(bounce, "Original-Recipient: %s", rcpt->dsn_orcpt);
+    } else if (NON_NULL_EMPTY(rcpt->orig_addr)) {
+	/* Fix 20140708: Don't send "utf-8" type with non-UTF8 address. */
+	post_mail_fprintf(bounce, "Original-Recipient: %s; %s",
+			  (bounce_info->sendopts & SMTPUTF8_FLAG_ALL)
+			  && IS_UTF8_ADDRESS(rcpt->orig_addr) ?
+			  "utf-8" : "rfc822", rcpt->orig_addr);
     }
     post_mail_fprintf(bounce, "Action: %s",
-		      IS_FAIL_TEMPLATE(bounce_info->template) ?
-		      "failed" : bounce_info->log_handle->dsn.action);
-    post_mail_fprintf(bounce, "Status: %s",
-		      bounce_info->log_handle->dsn.status);
-    if (bounce_info->log_handle->dsn.mtype
-	&& bounce_info->log_handle->dsn.mname)
+		      IS_FAILURE_TEMPLATE(bounce_info->template) ?
+		      "failed" : dsn->action);
+    post_mail_fprintf(bounce, "Status: %s", dsn->status);
+    if (NON_NULL_EMPTY(dsn->mtype) && NON_NULL_EMPTY(dsn->mname))
 	bounce_print_wrap(bounce, bounce_info, "Remote-MTA: %s; %s",
-			  bounce_info->log_handle->dsn.mtype,
-			  bounce_info->log_handle->dsn.mname);
-    if (bounce_info->log_handle->dsn.dtype
-	&& bounce_info->log_handle->dsn.dtext)
+			  dsn->mtype, dsn->mname);
+    if (NON_NULL_EMPTY(dsn->dtype) && NON_NULL_EMPTY(dsn->dtext))
 	bounce_print_wrap(bounce, bounce_info, "Diagnostic-Code: %s; %s",
-			  bounce_info->log_handle->dsn.dtype,
-			  bounce_info->log_handle->dsn.dtext);
+			  dsn->dtype, dsn->dtext);
     else
 	bounce_print_wrap(bounce, bounce_info, "Diagnostic-Code: X-%s; %s",
-			  bounce_info->mail_name,
-			  bounce_info->log_handle->dsn.reason);
+			  bounce_info->mail_name, dsn->reason);
 #if 0
-    post_mail_fprintf(bounce, "Last-Attempt-Date: %s",
-		      bounce_info->log_handle->log_time);
+    if (dsn->time > 0)
+	post_mail_fprintf(bounce, "Last-Attempt-Date: %s",
+			  mail_date(dsn->time));
 #endif
     if (IS_DELAY_TEMPLATE(bounce_info->template))
 	post_mail_fprintf(bounce, "Will-Retry-Until: %s",
@@ -650,6 +811,7 @@ int     bounce_recipient_dsn(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
 int     bounce_diagnostic_dsn(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
 			              int notify_filter)
 {
+    RECIPIENT *rcpt = &bounce_info->rcpt_buf->rcpt;
     int     count = 0;
 
     /*
@@ -663,12 +825,13 @@ int     bounce_diagnostic_dsn(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
      */
     if (bounce_info->log_handle == 0
 	|| bounce_log_rewind(bounce_info->log_handle)) {
-	if (IS_FAIL_TEMPLATE(bounce_info->template))
+	if (IS_FAILURE_TEMPLATE(bounce_info->template))
 	    count = 1;				/* XXX don't abort */
     } else {
-	while (bounce_log_read(bounce_info->log_handle) != 0) {
-	    if (bounce_info->log_handle->rcpt.dsn_notify == 0	/* compat */
-	    || (bounce_info->log_handle->rcpt.dsn_notify & notify_filter)) {
+	while (bounce_log_read(bounce_info->log_handle, bounce_info->rcpt_buf,
+			       bounce_info->dsn_buf) != 0) {
+	    if (rcpt->dsn_notify == 0		/* compat */
+		|| (rcpt->dsn_notify & notify_filter)) {
 		count++;
 		if (bounce_recipient_dsn(bounce, bounce_info) != 0)
 		    break;
@@ -700,7 +863,7 @@ int     bounce_original(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
      * MIME headers.
      */
 #define IS_UNDELIVERED_TEMPLATE(template) \
-        (IS_FAIL_TEMPLATE(template) || IS_DELAY_TEMPLATE(template))
+        (IS_FAILURE_TEMPLATE(template) || IS_DELAY_TEMPLATE(template))
 
     post_mail_fputs(bounce, "");
     post_mail_fprintf(bounce, "--%s", bounce_info->mime_boundary);
@@ -709,10 +872,16 @@ int     bounce_original(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
 		      "Undelivered " : "",
 		      headers_only == DSN_RET_HDRS ?
 		      "Message Headers" : "Message");
-    post_mail_fprintf(bounce, "Content-Type: %s",
-		      headers_only == DSN_RET_HDRS ?
-		      "text/rfc822-headers" : "message/rfc822");
-    if (bounce_info->mime_encoding)
+    /* Generate *global* only if the original requested SMTPUTF8 support. */
+    if (bounce_info->sendopts & SMTPUTF8_FLAG_REQUESTED)
+	post_mail_fprintf(bounce, "Content-Type: message/%s",
+			  headers_only == DSN_RET_HDRS ?
+			  "global-headers" : "global");
+    else
+	post_mail_fprintf(bounce, "Content-Type: %s",
+			  headers_only == DSN_RET_HDRS ?
+			  "text/rfc822-headers" : "message/rfc822");
+    if (NOT_7BIT_MIME(bounce_info))
 	post_mail_fprintf(bounce, "Content-Transfer-Encoding: %s",
 			  bounce_info->mime_encoding);
     post_mail_fputs(bounce, "");
@@ -726,6 +895,16 @@ int     bounce_original(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
 	return (vstream_ferror(bounce));
     }
 
+    /*
+     * XXX The cleanup server removes Return-Path: headers. This should be
+     * done only with mail that enters via a non-SMTP channel, but changing
+     * this now could break other software. Removing Return-Path: could break
+     * digital signatures, though this is unlikely. In any case,
+     * header_checks are more effective when the Return-Path: header is
+     * present, so we prepend one to the bounce message.
+     */
+    post_mail_fprintf(bounce, "Return-Path: <%s>", STR(bounce_info->sender));
+
     /*
      * Copy the original message contents. We're doing raw record output here
      * so that we don't throw away binary transparency yet.
@@ -758,22 +937,23 @@ int     bounce_original(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
 
 void    bounce_delrcpt(BOUNCE_INFO *bounce_info)
 {
+    RECIPIENT *rcpt = &bounce_info->rcpt_buf->rcpt;
+
     if (bounce_info->orig_fp != 0
 	&& bounce_info->log_handle != 0
 	&& bounce_log_rewind(bounce_info->log_handle) == 0)
-	while (bounce_log_read(bounce_info->log_handle) != 0)
-	    if (bounce_info->log_handle->rcpt.offset > 0)
-		deliver_completed(bounce_info->orig_fp,
-				  bounce_info->log_handle->rcpt.offset);
+	while (bounce_log_read(bounce_info->log_handle, bounce_info->rcpt_buf,
+			       bounce_info->dsn_buf) != 0)
+	    if (rcpt->offset > 0)
+		deliver_completed(bounce_info->orig_fp, rcpt->offset);
 }
 
 /* bounce_delrcpt_one - delete one recipient from original queue file */
 
 void    bounce_delrcpt_one(BOUNCE_INFO *bounce_info)
 {
-    if (bounce_info->orig_fp != 0
-	&& bounce_info->log_handle != 0
-	&& bounce_info->log_handle->rcpt.offset > 0)
-	deliver_completed(bounce_info->orig_fp,
-			  bounce_info->log_handle->rcpt.offset);
+    RECIPIENT *rcpt = &bounce_info->rcpt_buf->rcpt;
+
+    if (bounce_info->orig_fp != 0 && rcpt->offset > 0)
+	deliver_completed(bounce_info->orig_fp, rcpt->offset);
 }
