@@ -37,6 +37,14 @@
 /*	IBM T.J. Watson Research
 /*	P.O. Box 704
 /*	Yorktown Heights, NY 10598, USA
+/*
+/*	Wietse Venema
+/*	Google, Inc.
+/*	111 8th Avenue
+/*	New York, NY 10011, USA
+/*
+/*	Wietse Venema
+/*	porcupine.org
 /*--*/
 
 /* System library. */
@@ -44,6 +52,7 @@
 #include <sys_defs.h>
 #include <string.h>
 #include <stdlib.h>
+#include <stdio.h>			/* ssscanf() */
 #include <ctype.h>
 
 /* Utility library. */
@@ -66,6 +75,8 @@
 #include <mail_proto.h>
 #include <dsn_mask.h>
 #include <rec_attr_map.h>
+#include <smtputf8.h>
+#include <deliver_request.h>
 
 /* Application-specific. */
 
@@ -89,10 +100,12 @@ void    cleanup_envelope(CLEANUP_STATE *state, int type,
      * first, for backwards compatibility reasons.
      */
     cleanup_out_format(state, REC_TYPE_SIZE, REC_TYPE_SIZE_FORMAT,
-		       (REC_TYPE_SIZE_CAST1) 0,	/* content size */
+		       (REC_TYPE_SIZE_CAST1) 0,	/* extra offs - content offs */
 		       (REC_TYPE_SIZE_CAST2) 0,	/* content offset */
 		       (REC_TYPE_SIZE_CAST3) 0,	/* recipient count */
-		       (REC_TYPE_SIZE_CAST4) 0);	/* qmgr options */
+		       (REC_TYPE_SIZE_CAST4) 0,	/* qmgr options */
+		       (REC_TYPE_SIZE_CAST5) 0,	/* content length */
+		       (REC_TYPE_SIZE_CAST6) 0);	/* smtputf8 */
 
     /*
      * Pass control to the actual envelope processing routine.
@@ -145,15 +158,17 @@ static void cleanup_envelope_process(CLEANUP_STATE *state, int type,
 	return;
     }
 #endif
+
+    /*
+     * XXX We instantiate a MILTERS structure even when the filter count is
+     * zero (for example, all filters are in ACCEPT state, or the SMTP server
+     * sends a dummy MILTERS structure without any filters), otherwise the
+     * cleanup server would apply the non_smtpd_milters setting
+     * inappropriately.
+     */
     if (type == REC_TYPE_MILT_COUNT) {
 	/* Not part of queue file format. */
-	if (state->milters != 0) {
-	    msg_warn("%s: message rejected: too many milter instances",
-		     state->queue_id);
-	    state->errs |= CLEANUP_STAT_BAD;
-	    return;
-	}
-	if ((milter_count = atoi(buf)) > 0)
+	if ((milter_count = atoi(buf)) >= 0)
 	    cleanup_milter_receive(state, milter_count);
 	return;
     }
@@ -264,7 +279,7 @@ static void cleanup_envelope_process(CLEANUP_STATE *state, int type,
 	if (cleanup_milters != 0
 	    && state->milters == 0
 	    && CLEANUP_MILTER_OK(state))
-	    cleanup_milter_emul_rcpt(state, cleanup_milters, buf);
+	    cleanup_milter_emul_rcpt(state, cleanup_milters, state->recip);
 	myfree(state->orig_rcpt);
 	state->orig_rcpt = 0;
 	if (state->dsn_orcpt != 0) {
@@ -337,13 +352,20 @@ static void cleanup_envelope_process(CLEANUP_STATE *state, int type,
 
     /*
      * Initial envelope non-recipient record processing.
+     * 
+     * If this message was requeued with "postsuper -r", use their sender
+     * options flags, excluding flags derived from headers or envelopes.
+     * Those flags may be derived again, depending on Postfix configuration.
      */
     if (state->flags & CLEANUP_FLAG_INRCPT)
 	/* Tell qmgr that recipient records are mixed with other information. */
 	state->qmgr_opts |= QMGR_READ_FLAG_MIXED_RCPT_OTHER;
-    if (type == REC_TYPE_SIZE)
-	/* Use our own SIZE record instead. */
+    if (type == REC_TYPE_SIZE) {
+	/* Ignore their SIZE record, but keep the non-derived sender options. */
+	(void) sscanf(buf, "%*s $*s %*s %*s %*s %d", &state->sendopts);
+	state->sendopts &= ~SOPT_FLAG_DERIVED;
 	return;
+    }
     if (mapped_type == REC_TYPE_CTIME)
 	/* Use our own expiration time base record instead. */
 	return;
@@ -367,6 +389,8 @@ static void cleanup_envelope_process(CLEANUP_STATE *state, int type,
 	return;
     }
     if (type == REC_TYPE_FROM) {
+	off_t   after_sender_offs;
+
 	/* Allow only one instance. */
 	if (state->sender != 0) {
 	    msg_warn("%s: message rejected: multiple envelope sender records",
@@ -374,39 +398,38 @@ static void cleanup_envelope_process(CLEANUP_STATE *state, int type,
 	    state->errs |= CLEANUP_STAT_BAD;
 	    return;
 	}
-	cleanup_addr_sender(state, buf);
+	if (state->milters || cleanup_milters) {
+	    /* Remember the sender record offset. */
+	    if ((state->sender_pt_offset = vstream_ftell(state->dst)) < 0)
+		msg_fatal("%s: vstream_ftell %s: %m:", myname, cleanup_path);
+	}
+	after_sender_offs = cleanup_addr_sender(state, buf);
+	if (state->milters || cleanup_milters) {
+	    /* Remember the after-sender record offset. */
+	    state->sender_pt_target = after_sender_offs;
+	}
 	if (cleanup_milters != 0
 	    && state->milters == 0
 	    && CLEANUP_MILTER_OK(state))
-	    cleanup_milter_emul_mail(state, cleanup_milters, buf);
+	    cleanup_milter_emul_mail(state, cleanup_milters, state->sender);
 	return;
     }
     if (mapped_type == REC_TYPE_DSN_ENVID) {
-	/* Allow only one instance. */
-	if (state->dsn_envid != 0) {
-	    msg_warn("%s: message rejected: multiple DSN envelope ID records",
-		     state->queue_id);
-	    state->errs |= CLEANUP_STAT_BAD;
-	    return;
-	}
+	/* Don't break "postsuper -r" after Milter overrides ENVID. */
 	if (!allprint(mapped_buf)) {
 	    msg_warn("%s: message rejected: bad DSN envelope ID record",
 		     state->queue_id);
 	    state->errs |= CLEANUP_STAT_BAD;
 	    return;
 	}
+	if (state->dsn_envid != 0)
+	    myfree(state->dsn_envid);
 	state->dsn_envid = mystrdup(mapped_buf);
 	cleanup_out(state, type, buf, len);
 	return;
     }
     if (mapped_type == REC_TYPE_DSN_RET) {
-	/* Allow only one instance. */
-	if (state->dsn_ret != 0) {
-	    msg_warn("%s: message rejected: multiple DSN RET records",
-		     state->queue_id);
-	    state->errs |= CLEANUP_STAT_BAD;
-	    return;
-	}
+	/* Don't break "postsuper -r" after Milter overrides RET. */
 	if (!alldig(mapped_buf) || (junk = atoi(mapped_buf)) == 0
 	    || DSN_RET_OK(junk) == 0) {
 	    msg_warn("%s: message rejected: bad DSN RET record <%.200s>",
@@ -463,6 +486,16 @@ static void cleanup_envelope_process(CLEANUP_STATE *state, int type,
 		return;
 	    }
 	}
+	if (strcmp(attr_name, MAIL_ATTR_TRACE_FLAGS) == 0) {
+	    if (!alldig(attr_value)) {
+		msg_warn("%s: message rejected: bad TFLAG record <%.200s>",
+			 state->queue_id, buf);
+		state->errs |= CLEANUP_STAT_BAD;
+		return;
+	    }
+	    if (state->tflags == 0)
+		state->tflags = DEL_REQ_TRACE_FLAGS(atoi(attr_value));
+	}
 	nvtable_update(state->attr, attr_name, attr_value);
 	cleanup_out(state, type, buf, len);
 	return;
