@@ -603,8 +603,8 @@
 /* .PP
 /*	Available in Postfix version 3.2 and later:
 /* .IP "\fBtls_eecdh_auto_curves (see 'postconf -d' output)\fR"
-/*	The prioritized list of elliptic curves supported by the Postfix
-/*	SMTP client and server.
+/*	The prioritized list of elliptic curves, that should be enabled in the
+/*	Postfix SMTP client and server.
 /* .PP
 /*	Available in Postfix version 3.4 and later:
 /* .IP "\fBsmtp_tls_connection_reuse (no)\fR"
@@ -644,6 +644,10 @@
 /* .IP "\fBsmtp_tlsrpt_socket_name (empty)\fR"
 /*	The pathname of a UNIX-domain datagram socket that is managed
 /*	by a local TLSRPT reporting service.
+/* .IP "\fBsmtp_tlsrpt_skip_reused_handshakes (yes)\fR"
+/*	Do not report the TLSRPT status for TLS protocol handshakes
+/*	that reuse a previously-negotiated TLS session (there is no new
+/*	information to report).
 /* OBSOLETE STARTTLS CONTROLS
 /* .ad
 /* .fi
@@ -826,7 +830,7 @@
 /*	request before it is terminated by a built-in watchdog timer.
 /* .IP "\fBdelay_logging_resolution_limit (2)\fR"
 /*	The maximal number of digits after the decimal point when logging
-/*	sub-second delay values.
+/*	delay values.
 /* .IP "\fBdisable_dns_lookups (no)\fR"
 /*	Disable DNS lookups in the Postfix SMTP and LMTP clients.
 /* .IP "\fBinet_interfaces (all)\fR"
@@ -1154,6 +1158,7 @@ bool    var_ign_srv_lookup_err;
 bool    var_allow_srv_fallback;
 bool    var_smtp_tlsrpt_enable;
 char   *var_smtp_tlsrpt_sockname;
+bool    var_smtp_tlsrpt_skip_reused_hs;
 
  /* Special handling of 535 AUTH errors. */
 char   *var_smtp_sasl_auth_cache_name;
@@ -1407,7 +1412,23 @@ static void post_init(char *unused_name, char **argv)
 	var_disable_dns = (smtp_dns_support == SMTP_DNS_DISABLED);
     }
 
+#if !defined(USE_TLS) || !defined(USE_TLSRPT)
+    if (var_smtp_tlsrpt_enable)
+	msg_warn("TLSRPT is selected, but TLSRPT is not compiled in");
+#endif
 #ifdef USE_TLS
+#ifdef USE_TLSRPT
+    if (var_smtp_tlsrpt_enable) {
+	if (smtp_mode) {
+	    if (smtp_tlsrpt_post_jail(VAR_SMTP_TLSRPT_SOCKNAME,
+				      var_smtp_tlsrpt_sockname) < 0)
+		var_smtp_tlsrpt_enable = 0;
+	} else {
+	    msg_warn("TLSRPT support is not implemented for LMTP");
+	    var_smtp_tlsrpt_enable = 0;
+	}
+    }
+#endif						/* USE_TLSRPT */
     if (smtp_mode) {
 	smtp_tls_insecure_mx_policy =
 	    tls_level_lookup(var_smtp_tls_insecure_mx_policy);
@@ -1583,21 +1604,6 @@ static void pre_init(char *unused_name, char **unused_argv)
 			    mdalg = var_smtp_tls_fpt_dgst);
 	smtp_tls_list_init();
 	tls_dane_loglevel(VAR_LMTP_SMTP(TLS_LOGLEVEL), var_smtp_tls_loglevel);
-#ifdef USE_TLSRPT
-	if (var_smtp_tlsrpt_enable) {
-	    if (smtp_mode) {
-		if (smtp_tlsrpt_pre_jail(VAR_SMTP_TLSRPT_SOCKNAME,
-					 var_smtp_tlsrpt_sockname) < 0)
-		    var_smtp_tlsrpt_enable = 0;
-	    } else {
-		msg_warn("TLSRPT support is not implemented for LMTP");
-		var_smtp_tlsrpt_enable = 0;
-	    }
-	}
-#else						/* no USE_TLSRPT */
-	if (var_smtp_tlsrpt_enable)
-	    msg_warn("TLSRPT is selected, but TLSRPT is not compiled in");
-#endif						/* USE_TLSRPT */
 #else
 	msg_warn("TLS has been selected, but TLS support is not compiled in");
 #endif
