@@ -921,6 +921,7 @@ static void smtp_connect_inet(SMTP_STATE *state, const char *nexthop,
 	 */
 #ifdef USE_TLSRPT
 	if (smtp_mode && var_smtp_tlsrpt_enable
+	    && state->tls->level > TLS_LEV_NONE
 	    && !valid_hostaddr(domain, DONT_GRIPE))
 	    smtp_tlsrpt_create_wrapper(state, domain);
 	else
@@ -1092,8 +1093,14 @@ static void smtp_connect_inet(SMTP_STATE *state, const char *nexthop,
 		session->state = state;
 #ifdef USE_TLS
 		session->tls_nexthop = domain;
+
+		/*
+		 * Update TLSRPT state even if this is a reused SMTP
+		 * connection. If for some unlikely reason we must report a
+		 * problem, then we must report correct information.
+		 */
 #ifdef USE_TLSRPT
-		if (state->tlsrpt && state->tls->level > TLS_LEV_NONE) {
+		if (state->tlsrpt) {
 		    smtp_tlsrpt_set_tls_policy(state);
 		    smtp_tlsrpt_set_tcp_connection(state);
 		}
